# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging

from typing import Optional, Union

import pandera as pa

from hypothesis.strategies import DrawFn, SearchStrategy, composite

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_COLUMNS_KEY,
    CUSTOM_DATA_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_TYPE_KEY,
    PANDERA_SCHEMA_KEY,
    PYSPARK_ARRAY_TYPE,
    PYSPARK_BINARY_TYPE,
    PYSPARK_DATE_TYPE,
    PYSPARK_STRING_TYPE,
    PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES,
)
from snowflake.hypothesis_snowpark.custom_strategies import update_pandas_df_strategy
from snowflake.hypothesis_snowpark.logging_utils import log
from snowflake.hypothesis_snowpark.strategies_utils import (
    apply_custom_null_values,
    generate_snowpark_dataframe,
    generate_snowpark_schema,
    load_json_schema,
    replace_surrogate_chars,
)
from snowflake.hypothesis_snowpark.telemetry.telemetry import report_telemetry
from snowflake.snowpark import DataFrame, Session


LOGGER = logging.getLogger(__name__)


@report_telemetry(params_list=["schema"])
@log
def dataframe_strategy(
    schema: Union[str, pa.DataFrameSchema], session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a given schema.

    Args:
        schema: A schema defining the columns, data types and checks that the generated DataFrame
                should satisfy. This can be a path to a JSON schema file generated by the
                :func:`snowflake.snowpark_checkpoints_collector.collect_dataframe_checkpoint`
                function when the collection mode is set to `SCHEMA`, or a Pandera DataFrameSchema
                object.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate for each DataFrame. If not specified, the strategy will
              generate DataFrames of different sizes.

    Examples:
        Generate a Snowpark DataFrame from a JSON schema file:

        >>> from hypothesis import given
        >>> from snowflake.hypothesis_snowpark import dataframe_strategy
        >>> from snowflake.snowpark import DataFrame, Session
            <BLANKLINE>
        >>> @given(
        ...     df=dataframe_strategy(
        ...         schema="path/to/schema.json",
        ...         session=Session.builder.getOrCreate(),
        ...         size=10,
        ...     )
        ... )
        >>> def test_my_function(df: DataFrame):
        ...     ...

        Generate a Snowpark DataFrame from a Pandera DataFrameSchema object:

        >>> import pandera as pa
        >>> from hypothesis import given
        >>> from snowflake.hypothesis_snowpark import dataframe_strategy
        >>> from snowflake.snowpark import DataFrame, Session
            <BLANKLINE>
        >>> @given(
        ...    df=dataframe_strategy(
        ...        schema=pa.DataFrameSchema(
        ...            {
        ...                "A": pa.Column(pa.Int, checks=pa.Check.in_range(0, 10)),
        ...                "B": pa.Column(pa.Bool),
        ...            }
        ...        ),
        ...        session=Session.builder.getOrCreate(),
        ...        size=10,
        ...    )
        ... )
        >>> def test_my_function(df: DataFrame):
        ...     ...

        You can control aspects like the maximum number of test cases, the deadline for each test
        execution, verbosity levels and many others using the Hypothesis @settings decorator.

        >>> from datetime import timedelta
        >>> from hypothesis import given, settings
        >>> from snowflake.hypothesis_snowpark import dataframe_strategy
        >>> from snowflake.snowpark import DataFrame, Session
            <BLANKLINE>
        >>> @given(
        ...     df=dataframe_strategy(
        ...         schema="path/to/schema.json",
        ...         session=Session.builder.getOrCreate(),
        ...         size=10,
        ...     )
        ... )
        >>> @settings(
        ...     deadline=timedelta(milliseconds=800),
        ...     max_examples=25,
        ... )
        >>> def test_my_function(df: DataFrame):
        ...     ...

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    if not session:
        raise ValueError("Session cannot be None.")

    if isinstance(schema, pa.DataFrameSchema):
        return _dataframe_strategy_from_object_schema(schema, session, size)

    if isinstance(schema, str) and schema.endswith(".json"):
        return _dataframe_strategy_from_json_schema(schema, session, size)

    raise ValueError(
        "Schema must be a path to a JSON schema file or a Pandera DataFrameSchema object."
    )


def _dataframe_strategy_from_object_schema(
    schema: pa.DataFrameSchema, session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a Pandera DataFrameSchema object.

    Args:
        schema: The Pandera DataFrameSchema object.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate.

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    LOGGER.info(
        "Starting to create a Hypothesis strategy from the given Pandera DataFrameSchema object"
    )
    original_schema = copy.deepcopy(schema)
    str_columns = []

    for column in schema.columns.values():
        if isinstance(column.dtype, pa.engines.pandas_engine.Date):
            # Data generation for date type is currently unsupported by Pandera.
            # As a workaround, we can change the data type to pa.DateTime to
            # avoid an exception and let Snowpark handle the conversion to Date.
            LOGGER.warning(
                "Data generation for Date type is not supported; converting column '%s' to Timestamp",
                column.name,
            )
            column.dtype = pa.DateTime
        elif isinstance(column.dtype, pa.String):
            str_columns.append(column.name)

    @composite
    @log
    def _dataframe_strategy(draw: DrawFn) -> DataFrame:
        pandas_strategy = schema.strategy(size=size)
        pandas_df = draw(pandas_strategy)
        LOGGER.info("Generated a Pandas DataFrame with shape: %s", pandas_df.shape)
        pandas_df = replace_surrogate_chars(pandas_df, str_columns)
        snowpark_schema = generate_snowpark_schema(original_schema)
        snowpark_df = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
        return snowpark_df

    LOGGER.info(
        "Successfully created a Hypothesis strategy from the given Pandera DataFrameSchema object"
    )
    return _dataframe_strategy()


def _dataframe_strategy_from_json_schema(
    schema: str, session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a JSON schema.

    Args:
        schema: The path to the JSON schema file.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate.

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    LOGGER.info(
        "Starting to create a Hypothesis strategy from the given JSON schema file"
    )
    json_schema_dict = load_json_schema(schema)
    pandera_schema = json_schema_dict.get(PANDERA_SCHEMA_KEY)
    custom_data = json_schema_dict.get(CUSTOM_DATA_KEY)

    if not (pandera_schema and custom_data):
        raise ValueError(
            f"Invalid JSON schema. The JSON schema must contain '{PANDERA_SCHEMA_KEY}' and '{CUSTOM_DATA_KEY}' keys."
        )

    custom_data_columns = custom_data.get(CUSTOM_DATA_COLUMNS_KEY, [])
    not_supported_columns, columns_with_custom_strategy, str_columns = [], [], []

    for column in custom_data_columns:
        dtype = column.get(CUSTOM_DATA_TYPE_KEY)
        if dtype not in PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES:
            not_supported_columns.append(column)
        elif dtype in (PYSPARK_ARRAY_TYPE, PYSPARK_BINARY_TYPE, PYSPARK_DATE_TYPE):
            columns_with_custom_strategy.append(column)
        elif dtype == PYSPARK_STRING_TYPE:
            str_columns.append(column.get(CUSTOM_DATA_NAME_KEY))

    if not_supported_columns:
        raise ValueError(
            f"The following data types are not supported by the Snowpark DataFrame strategy: "
            f"{[column.get(CUSTOM_DATA_TYPE_KEY) for column in not_supported_columns]}"
        )

    LOGGER.info("Creating a Pandera DataFrameSchema object from the given schema")
    df_schema = pa.DataFrameSchema.from_json(json.dumps(pandera_schema))
    LOGGER.debug("Loaded Pandera DataFrameSchema: %s", repr(df_schema))
    df_schema = _process_dataframe_schema(df_schema)

    @composite
    @log
    def _dataframe_strategy(draw: DrawFn) -> DataFrame:
        pandas_strategy = df_schema.strategy(size=size)
        pandas_df = draw(pandas_strategy)
        LOGGER.info("Generated a Pandas DataFrame with shape: %s", pandas_df.shape)
        pandas_df = draw(
            update_pandas_df_strategy(pandas_df, columns_with_custom_strategy)
        )
        pandas_df = apply_custom_null_values(pandas_df, custom_data)
        pandas_df = replace_surrogate_chars(pandas_df, str_columns)
        snowpark_schema = generate_snowpark_schema(df_schema, custom_data)
        snowpark_df = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
        return snowpark_df

    LOGGER.info(
        "Successfully created a Hypothesis strategy from the given JSON schema file"
    )
    return _dataframe_strategy()


def _process_dataframe_schema(df_schema: pa.DataFrameSchema) -> pa.DataFrameSchema:
    df_schema_copy = copy.copy(df_schema)

    for column_obj in df_schema_copy.columns.values():
        if isinstance(column_obj.dtype, pa.engines.pandas_engine.Date):
            # Data generation for date type is currently unsupported by Pandera. As a workaround,
            # we can change the data type to any supported type to avoid an exception and manually
            # generate the dates.
            LOGGER.warning(
                "Data generation for Date type is not supported; converting column '%s' to Timestamp",
                column_obj.name,
            )
            column_obj.dtype = pa.DateTime

    return df_schema_copy



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Wrapper around pandera which logs to snowflake

import logging

from typing import Any, Optional, Union, cast

from pandas import DataFrame as PandasDataFrame
from pandera import Check, DataFrameModel, DataFrameSchema
from pandera.errors import SchemaError, SchemaErrors

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.constants import (
    FAIL_STATUS,
    PASS_STATUS,
    CheckpointMode,
)
from snowflake.snowpark_checkpoints.utils.extra_config import is_checkpoint_enabled
from snowflake.snowpark_checkpoints.utils.logging_utils import log
from snowflake.snowpark_checkpoints.utils.pandera_check_manager import (
    PanderaCheckManager,
)
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _check_compare_data,
    _generate_schema,
    _process_sampling,
    _replace_special_characters,
    _update_validation_result,
)


LOGGER = logging.getLogger(__name__)


@log
def validate_dataframe_checkpoint(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    mode: Optional[CheckpointMode] = CheckpointMode.SCHEMA,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a Snowpark DataFrame against a specified checkpoint.

    Args:
        df (SnowparkDataFrame): The DataFrame to validate.
        checkpoint_name (str): The name of the checkpoint to validate against.
        job_context (SnowparkJobContext, optional): The job context for the validation. Required for PARQUET mode.
        mode (CheckpointMode): The mode of validation (e.g., SCHEMA, PARQUET). Defaults to SCHEMA.
        custom_checks (Optional[dict[Any, Any]], optional): Custom checks to apply during validation.
        skip_checks (Optional[dict[Any, Any]], optional): Checks to skip during validation.
        sample_frac (Optional[float], optional): Fraction of the DataFrame to sample for validation. Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample for validation.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy to use for sampling.
            Defaults to RANDOM_SAMPLE.
        output_path (Optional[str], optional): The output path for the validation results.

    Returns:
        Union[tuple[bool, PandasDataFrame], None]: A tuple containing a boolean indicating success
        and a Pandas DataFrame with validation results, or None if validation is not applicable.

    Raises:
        ValueError: If an invalid validation mode is provided or if job_context is None for PARQUET mode.

    """
    checkpoint_name = _replace_special_characters(checkpoint_name)

    if not is_checkpoint_enabled(checkpoint_name):
        LOGGER.warning(
            "Checkpoint '%s' is disabled. Skipping DataFrame checkpoint validation.",
            checkpoint_name,
        )
        return None

    LOGGER.info(
        "Starting DataFrame checkpoint validation for checkpoint '%s'", checkpoint_name
    )

    if mode == CheckpointMode.SCHEMA:
        result = _check_dataframe_schema_file(
            df,
            checkpoint_name,
            job_context,
            custom_checks,
            skip_checks,
            sample_frac,
            sample_number,
            sampling_strategy,
            output_path,
        )
        return result

    if mode == CheckpointMode.DATAFRAME:
        if job_context is None:
            raise ValueError(
                "No job context provided. Please provide one when using DataFrame mode validation."
            )
        _check_compare_data(df, job_context, checkpoint_name, output_path)
        return None

    raise ValueError(
        (
            "Invalid validation mode. "
            "Please use 1 for schema validation or 2 for full data validation."
        ),
    )


def _check_dataframe_schema_file(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:
    """Generate and checks the schema for a given DataFrame based on a checkpoint name.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        checkpoint_name (str): The name of the checkpoint to retrieve the schema.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        tuple[bool, PanderaDataFrame]: A tuple containing the validity flag and the Pandera DataFrame.

    """
    if df is None:
        raise ValueError("DataFrame is required")

    if checkpoint_name is None:
        raise ValueError("Checkpoint name is required")

    schema = _generate_schema(checkpoint_name, output_path)

    return _check_dataframe_schema(
        df,
        schema,
        checkpoint_name,
        job_context,
        custom_checks,
        skip_checks,
        sample_frac,
        sample_number,
        sampling_strategy,
        output_path,
    )


@log
def check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a DataFrame against a given Pandera schema using sampling techniques.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (str, optional): The name of the checkpoint to retrieve the schema.
            Defaults to None.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        Union[tuple[bool, PandasDataFrame]|None]: A tuple containing the validity flag and the Pandas DataFrame.
        If the validation for that checkpoint is disabled it returns None.

    """
    checkpoint_name = _replace_special_characters(checkpoint_name)
    LOGGER.info(
        "Starting DataFrame schema validation for checkpoint '%s'", checkpoint_name
    )

    if df is None:
        raise ValueError("DataFrame is required")

    if pandera_schema is None:
        raise ValueError("Schema is required")

    if not is_checkpoint_enabled(checkpoint_name):
        LOGGER.warning(
            "Checkpoint '%s' is disabled. Skipping DataFrame schema validation.",
            checkpoint_name,
        )
        return None

    return _check_dataframe_schema(
        df,
        pandera_schema,
        checkpoint_name,
        job_context,
        custom_checks,
        skip_checks,
        sample_frac,
        sample_number,
        sampling_strategy,
        output_path,
    )


@report_telemetry(
    params_list=["pandera_schema"],
    return_indexes=[(STATUS_KEY, 0)],
    multiple_return=True,
)
def _check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: SnowparkJobContext = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:

    pandera_check_manager = PanderaCheckManager(checkpoint_name, pandera_schema)
    pandera_check_manager.skip_checks_on_schema(skip_checks)
    pandera_check_manager.add_custom_checks(custom_checks)

    pandera_schema_upper, sample_df = _process_sampling(
        df, pandera_schema, job_context, sample_frac, sample_number, sampling_strategy
    )
    is_valid, validation_result = _validate(pandera_schema_upper, sample_df)
    if is_valid:
        LOGGER.info(
            "DataFrame schema validation passed for checkpoint '%s'",
            checkpoint_name,
        )
        if job_context is not None:
            job_context._mark_pass(checkpoint_name)
        else:
            LOGGER.warning(
                "No job context provided. Skipping result recording into Snowflake.",
            )
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
    else:
        LOGGER.error(
            "DataFrame schema validation failed for checkpoint '%s'",
            checkpoint_name,
        )
        _update_validation_result(checkpoint_name, FAIL_STATUS, output_path)
        raise SchemaValidationError(
            "Snowpark DataFrame schema validation error",
            job_context,
            checkpoint_name,
            validation_result,
        )

    return (is_valid, validation_result)


@report_telemetry(params_list=["pandera_schema"])
@log
def check_output_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate to validate the schema of the output of a Snowpark function.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.

    """

    def check_output_with_decorator(snowpark_fn):
        """Decorate to validate the schema of the output of a Snowpark function.

        Args:
            snowpark_fn (function): The Snowpark function to validate.

        Returns:
            function: The decorated function.

        """

        @log(log_args=False)
        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the output of a Snowpark function.

            Args:
                *args: The arguments to the Snowpark function.
                **kwargs: The keyword arguments to the Snowpark function.

            Returns:
                Any: The result of the Snowpark function.

            """
            _checkpoint_name = checkpoint_name
            if checkpoint_name is None:
                LOGGER.warning(
                    (
                        "No checkpoint name provided for output schema validation. "
                        "Using '%s' as the checkpoint name.",
                    ),
                    snowpark_fn.__name__,
                )
                _checkpoint_name = snowpark_fn.__name__
            _checkpoint_name = _replace_special_characters(_checkpoint_name)
            LOGGER.info(
                "Starting output schema validation for Snowpark function '%s' and checkpoint '%s'",
                snowpark_fn.__name__,
                _checkpoint_name,
            )

            # Run the sampled data in snowpark
            LOGGER.info("Running the Snowpark function '%s'", snowpark_fn.__name__)
            snowpark_results = snowpark_fn(*args, **kwargs)
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args([snowpark_results])
            pandas_sample_args = sampler.get_sampled_pandas_args()

            is_valid, validation_result = _validate(
                pandera_schema, pandas_sample_args[0]
            )
            if is_valid:
                LOGGER.info(
                    "Output schema validation passed for Snowpark function '%s' and checkpoint '%s'",
                    snowpark_fn.__name__,
                    _checkpoint_name,
                )
                if job_context is not None:
                    job_context._mark_pass(_checkpoint_name)
                else:
                    LOGGER.warning(
                        "No job context provided. Skipping result recording into Snowflake.",
                    )
                _update_validation_result(_checkpoint_name, PASS_STATUS, output_path)
            else:
                LOGGER.error(
                    "Output schema validation failed for Snowpark function '%s' and checkpoint '%s'",
                    snowpark_fn.__name__,
                    _checkpoint_name,
                )
                _update_validation_result(_checkpoint_name, FAIL_STATUS, output_path)
                raise SchemaValidationError(
                    "Snowpark output schema validation error",
                    job_context,
                    _checkpoint_name,
                    validation_result,
                )
            return snowpark_results

        return wrapper

    return check_output_with_decorator


@report_telemetry(params_list=["pandera_schema"])
@log
def check_input_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate factory for validating input DataFrame schemas before function execution.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.


    """

    def check_input_with_decorator(snowpark_fn):
        """Decorate that validates input schemas for the decorated function.

        Args:
            snowpark_fn (Callable): The function to be decorated with input schema validation.

        Raises:
            SchemaValidationError: If input data fails schema validation.

        Returns:
            Callable: A wrapper function that performs schema validation before executing the original function.

        """

        @log(log_args=False)
        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the input of a Snowpark function.

            Raises:
                SchemaValidationError: If any input DataFrame fails schema validation.

            Returns:
                Any: The result of the original function after input validation.

            """
            _checkpoint_name = checkpoint_name
            if checkpoint_name is None:
                LOGGER.warning(
                    (
                        "No checkpoint name provided for input schema validation. "
                        "Using '%s' as the checkpoint name."
                    ),
                    snowpark_fn.__name__,
                )
                _checkpoint_name = snowpark_fn.__name__
            _checkpoint_name = _replace_special_characters(_checkpoint_name)
            LOGGER.info(
                "Starting input schema validation for Snowpark function '%s' and checkpoint '%s'",
                snowpark_fn.__name__,
                _checkpoint_name,
            )

            # Run the sampled data in snowpark
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args(args)
            pandas_sample_args = sampler.get_sampled_pandas_args()

            LOGGER.info(
                "Validating %s input argument(s) against a Pandera schema",
                len(pandas_sample_args),
            )
            # Raises SchemaError on validation issues
            for index, arg in enumerate(pandas_sample_args, start=1):
                if not isinstance(arg, PandasDataFrame):
                    LOGGER.info(
                        "Arg %s: Skipping schema validation for non-DataFrame argument",
                        index,
                    )
                    continue

                is_valid, validation_result = _validate(
                    pandera_schema,
                    arg,
                )
                if is_valid:
                    LOGGER.info(
                        "Arg %s: Input schema validation passed",
                        index,
                    )
                    if job_context is not None:
                        job_context._mark_pass(
                            _checkpoint_name,
                        )
                    _update_validation_result(
                        _checkpoint_name,
                        PASS_STATUS,
                        output_path,
                    )
                else:
                    LOGGER.error(
                        "Arg %s: Input schema validation failed",
                        index,
                    )
                    _update_validation_result(
                        _checkpoint_name,
                        FAIL_STATUS,
                        output_path,
                    )
                    raise SchemaValidationError(
                        "Snowpark input schema validation error",
                        job_context,
                        _checkpoint_name,
                        validation_result,
                    )
            return snowpark_fn(*args, **kwargs)

        return wrapper

    return check_input_with_decorator


def _validate(
    schema: Union[type[DataFrameModel], DataFrameSchema],
    df: PandasDataFrame,
    lazy: bool = True,
) -> tuple[bool, PandasDataFrame]:
    if not isinstance(schema, DataFrameSchema):
        schema = schema.to_schema()
    is_valid = True
    try:
        df = schema.validate(df, lazy=lazy)
    except (SchemaErrors, SchemaError) as schema_errors:
        df = cast(PandasDataFrame, schema_errors.failure_cases)
        is_valid = False
    return is_valid, df



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from typing import Optional

from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.alias_generators import to_camel

from snowflake.snowpark_checkpoints_configuration import checkpoint_name_utils


LOGGER = logging.getLogger(__name__)


class Checkpoint(BaseModel):

    """Checkpoint model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    """

    name: str
    mode: int = 1
    function: Optional[str] = None
    df: Optional[str] = None
    sample: Optional[float] = None
    file: Optional[str] = None
    location: int = -1
    enabled: bool = True

    @field_validator("name", mode="before")
    @classmethod
    def normalize(cls, name: str) -> str:
        LOGGER.debug("Normalizing checkpoint name: '%s'", name)
        normalized_name = checkpoint_name_utils.normalize_checkpoint_name(name)
        LOGGER.debug("Checkpoint name was normalized to: '%s'", normalized_name)
        is_valid_checkpoint_name = checkpoint_name_utils.is_valid_checkpoint_name(
            normalized_name
        )
        if not is_valid_checkpoint_name:
            error_msg = (
                f"Invalid checkpoint name: {name} in checkpoints.json file. "
                f"Checkpoint names must only contain alphanumeric characters, underscores and dollar signs."
            )
            LOGGER.error(error_msg)
            raise Exception(error_msg)

        return normalized_name


class Pipeline(BaseModel):

    """Pipeline model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    entry_point: str
    checkpoints: list[Checkpoint]


class Checkpoints(BaseModel):

    """Checkpoints model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    Returns:
        Checkpoints: An instance of the Checkpoints class

    """

    type: str
    pipelines: list[Pipeline]

    # this dictionary holds the unpacked checkpoints from the different pipelines.
    _checkpoints = {}

    def _build_checkpoints_dict(self):
        LOGGER.debug("Building checkpoints dictionary from pipelines.")
        for pipeline in self.pipelines:
            LOGGER.debug("Processing pipeline: '%s'", pipeline.entry_point)
            for checkpoint in pipeline.checkpoints:
                LOGGER.debug("Adding checkpoint: %s", checkpoint)
                self._checkpoints[checkpoint.name] = checkpoint

    def get_check_point(self, checkpoint_name: str) -> Checkpoint:
        """Get a checkpoint by its name.

        Args:
            checkpoint_name (str): The name of the checkpoint.

        Returns:
            Checkpoint: The checkpoint object if found, otherwise a new Checkpoint object
            with the name set to the checkpoint_id.

        """
        LOGGER.info("Fetching checkpoint: '%s'", checkpoint_name)
        if not self._checkpoints:
            LOGGER.debug("Checkpoints dictionary is empty, building it...")
            self._build_checkpoints_dict()

        checkpoint = self._checkpoints.get(checkpoint_name)
        if len(self._checkpoints) == 0:
            LOGGER.info(
                "No checkpoints found, creating a new enabled checkpoint with name: '%s'",
                checkpoint_name,
            )
            checkpoint = Checkpoint(name=checkpoint_name, enabled=True)
        elif checkpoint is None:
            LOGGER.info(
                "Checkpoint not found, creating a new disabled checkpoint with name: '%s'",
                checkpoint_name,
            )
            checkpoint = Checkpoint(name=checkpoint_name, enabled=False)

        LOGGER.debug("Returning checkpoint: %s", checkpoint)
        return checkpoint

    def add_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Add a checkpoint to the checkpoints' dictionary.

        Args:
            checkpoint (Checkpoint): The checkpoint object to add

        """
        LOGGER.debug("Adding checkpoint: %s", checkpoint)
        self._checkpoints[checkpoint.name] = checkpoint
        LOGGER.info("Checkpoint '%s' added successfully.", checkpoint.name)



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pydantic import BaseModel


class ValidationResult(BaseModel):

    """ValidationResult represents the result of a validation checkpoint.

    Attributes:
        result (str): The result of the validation.
        timestamp (datetime): The timestamp when the validation was performed.
        file (str): The file where the validation checkpoint is located.
        line_of_code (int): The line number in the file where the validation checkpoint is located.
        checkpoint_name (str): The name of the validation checkpoint.

    """

    result: str
    timestamp: str
    file: str
    line_of_code: int
    checkpoint_name: str


class ValidationResults(BaseModel):

    """ValidationResults is a model that holds a list of validation results.

    Attributes:
        results (list[ValidationResult]): A list of validation results.

    """

    results: list[ValidationResult]



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import json
import logging
import os
import re

from datetime import datetime
from typing import Any, Optional

import numpy as np

from pandera import DataFrameSchema

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.constants import (
    CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME,
    CHECKPOINT_TABLE_NAME_FORMAT,
    COLUMNS_KEY,
    DATAFRAME_CUSTOM_DATA_KEY,
    DATAFRAME_EXECUTION_MODE,
    DATAFRAME_PANDERA_SCHEMA_KEY,
    DEFAULT_KEY,
    EXCEPT_HASH_AGG_QUERY,
    FAIL_STATUS,
    PASS_STATUS,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
)
from snowflake.snowpark_checkpoints.utils.extra_config import (
    get_checkpoint_file,
)
from snowflake.snowpark_checkpoints.utils.pandera_check_manager import (
    PanderaCheckManager,
)
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.validation_result_metadata import (
    ValidationResultsMetadata,
)
from snowflake.snowpark_checkpoints.validation_results import ValidationResult


LOGGER = logging.getLogger(__name__)


def _replace_special_characters(checkpoint_name: str) -> str:
    """Replace special characters in the checkpoint name with underscores.

    Args:
        checkpoint_name (str): The checkpoint name to process.

    Returns:
        str: The checkpoint name with special characters replaced by underscores.

    """
    regex = r"^[a-zA-Z_\s-][a-zA-Z0-9$_\s-]*$"
    if not bool(re.match(regex, checkpoint_name)):
        raise ValueError(
            f"Invalid checkpoint name: {checkpoint_name}",
            "Checkpoint name must contain only alphanumeric characters, hyphens, underscores and dollar signs.",
        )
    return re.sub(r"[\s-]", "_", checkpoint_name)


def _process_sampling(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    job_context: Optional[SnowparkJobContext] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
):
    """Process a Snowpark DataFrame by sampling it according to the specified parameters.

    Adjusts the column casing of the provided Pandera schema to uppercase.

    Args:
        df (SnowparkDataFrame): The Snowpark DataFrame to be sampled.
        pandera_schema (DataFrameSchema): The Pandera schema to validate the DataFrame.
        job_context (SnowparkJobContext, optional): The job context for the sampling operation.
            Defaults to None.
        sample_frac (Optional[float], optional): The fraction of rows to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): The number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): The strategy to use for sampling.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.

    Returns:
        Tuple[DataFrameSchema, pd.DataFrame]: A tuple containing the adjusted Pandera schema with uppercase column names
        and the sampled pandas DataFrame.

    """
    sampler = SamplingAdapter(
        job_context, sample_frac, sample_number, sampling_strategy
    )
    sampler.process_args([df])

    # fix up the column casing
    pandera_schema_upper = pandera_schema
    new_columns: dict[Any, Any] = {}

    for col in pandera_schema.columns:
        new_columns[col.upper()] = pandera_schema.columns[col]

    pandera_schema_upper = pandera_schema_upper.remove_columns(pandera_schema.columns)
    pandera_schema_upper = pandera_schema_upper.add_columns(new_columns)

    sample_df = sampler.get_sampled_pandas_args()[0]
    sample_df.index = np.ones(sample_df.count().iloc[0])

    return pandera_schema_upper, sample_df


def _generate_schema(
    checkpoint_name: str, output_path: Optional[str] = None
) -> DataFrameSchema:
    """Generate a DataFrameSchema based on the checkpoint name provided.

    This function reads a JSON file corresponding to the checkpoint name,
    extracts schema information, and constructs a DataFrameSchema object.
    It also adds custom checks for numeric and boolean types if specified
    in the JSON file.

    Args:
        checkpoint_name (str): The name of the checkpoint used to locate
                               the JSON file containing schema information.
        output_path (str): The path to the output directory.

        DataFrameSchema: A schema object representing the structure and
                         constraints of the DataFrame.
                         constraints of the DataFrame.

    """
    LOGGER.info(
        "Generating Pandera DataFrameSchema for checkpoint: '%s'", checkpoint_name
    )
    current_directory_path = output_path if output_path else os.getcwd()

    output_directory_path = os.path.join(
        current_directory_path, SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME
    )

    if not os.path.exists(output_directory_path):
        raise ValueError(
            """Output directory snowpark-checkpoints-output does not exist.
Please run the Snowpark checkpoint collector first."""
        )

    checkpoint_schema_file_path = os.path.join(
        output_directory_path,
        CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME.format(checkpoint_name),
    )

    if not os.path.exists(checkpoint_schema_file_path):
        raise ValueError(
            f"Checkpoint {checkpoint_name} JSON file not found. Please run the Snowpark checkpoint collector first."
        )

    LOGGER.info("Reading schema from file: '%s'", checkpoint_schema_file_path)
    with open(checkpoint_schema_file_path) as schema_file:
        checkpoint_schema_config = json.load(schema_file)

    if DATAFRAME_PANDERA_SCHEMA_KEY not in checkpoint_schema_config:
        raise ValueError(
            f"Pandera schema not found in the JSON file for checkpoint: {checkpoint_name}"
        )

    schema_dict = checkpoint_schema_config.get(DATAFRAME_PANDERA_SCHEMA_KEY)
    schema_dict_str = json.dumps(schema_dict)
    schema = DataFrameSchema.from_json(schema_dict_str)

    if DATAFRAME_CUSTOM_DATA_KEY not in checkpoint_schema_config:
        LOGGER.info(
            "No custom data found in the JSON file for checkpoint: '%s'",
            checkpoint_name,
        )
        return schema

    custom_data = checkpoint_schema_config.get(DATAFRAME_CUSTOM_DATA_KEY)

    if COLUMNS_KEY not in custom_data:
        raise ValueError(
            f"Columns not found in the JSON file for checkpoint: {checkpoint_name}"
        )

    pandera_check_manager = PanderaCheckManager(
        checkpoint_name=checkpoint_name, schema=schema
    )
    schema = pandera_check_manager.proccess_checks(custom_data)

    return schema


def _check_compare_data(
    df: SnowparkDataFrame,
    job_context: Optional[SnowparkJobContext],
    checkpoint_name: str,
    output_path: Optional[str] = None,
):
    """Compare the data in the provided Snowpark DataFrame with the data in a checkpoint table.

    This function writes the provided DataFrame to a table and compares it with an existing checkpoint table
    using a hash aggregation query. If there is a data mismatch, it marks the job context as failed and raises a
    SchemaValidationError. If the data matches, it marks the job context as passed.

    Args:
        df (SnowparkDataFrame): The Snowpark DataFrame to compare.
        job_context (Optional[SnowparkJobContext]): The job context containing the Snowpark session and job state.
        checkpoint_name (str): The name of the checkpoint table to compare against.
        output_path (Optional[str]): The path to the output directory.

    Raises:
        SchemaValidationError: If there is a data mismatch between the DataFrame and the checkpoint table.

    """
    _, err = _compare_data(df, job_context, checkpoint_name, output_path)
    if err is not None:
        raise err


@report_telemetry(
    params_list=["df"], return_indexes=[(STATUS_KEY, 0)], multiple_return=True
)
def _compare_data(
    df: SnowparkDataFrame,
    job_context: Optional[SnowparkJobContext],
    checkpoint_name: str,
    output_path: Optional[str] = None,
) -> tuple[bool, Optional[SchemaValidationError]]:
    """Compare the data in the provided Snowpark DataFrame with the data in a checkpoint table.

    This function writes the provided DataFrame to a table and compares it with an existing checkpoint table
    using a hash aggregation query. If there is a data mismatch, it marks the job context as failed and raises a
    SchemaValidationError. If the data matches, it marks the job context as passed.

    Args:
        df (SnowparkDataFrame): The Snowpark DataFrame to compare.
        job_context (Optional[SnowparkJobContext]): The job context containing the Snowpark session and job state.
        checkpoint_name (str): The name of the checkpoint table to compare against.
        output_path (Optional[str]): The path to the output directory.

    Returns:
        Tuple[bool, Optional[SchemaValidationError]]: A tuple containing a boolean indicating if the data matches
        and an optional SchemaValidationError if there is a data mismatch.

    Raises:
        SchemaValidationError: If there is a data mismatch between the DataFrame and the checkpoint table.

    """
    new_table_name = CHECKPOINT_TABLE_NAME_FORMAT.format(checkpoint_name)
    LOGGER.info(
        "Writing Snowpark DataFrame to table: '%s' for checkpoint: '%s'",
        new_table_name,
        checkpoint_name,
    )
    df.write.save_as_table(table_name=new_table_name, mode="overwrite")

    LOGGER.info(
        "Comparing DataFrame to checkpoint table: '%s' for checkpoint: '%s'",
        new_table_name,
        checkpoint_name,
    )
    expect_df = job_context.snowpark_session.sql(
        EXCEPT_HASH_AGG_QUERY, [checkpoint_name, new_table_name]
    )

    if expect_df.count() != 0:
        error_message = f"Data mismatch for checkpoint {checkpoint_name}"
        job_context._mark_fail(
            error_message,
            checkpoint_name,
            df,
            DATAFRAME_EXECUTION_MODE,
        )
        _update_validation_result(
            checkpoint_name,
            FAIL_STATUS,
            output_path,
        )
        return False, SchemaValidationError(
            error_message,
            job_context,
            checkpoint_name,
            df,
        )
    else:
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        job_context._mark_pass(checkpoint_name, DATAFRAME_EXECUTION_MODE)
        return True, None


def _find_frame_in(stack: list[inspect.FrameInfo]) -> tuple:
    """Find a specific frame in the provided stack trace.

    This function searches through the provided stack trace to find a frame that matches
    certain criteria. It looks for frames where the function name is "wrapper" or where
    the code context matches specific regular expressions.

    Args:
        stack (list[inspect.FrameInfo]): A list of frame information objects representing
                                         the current stack trace.

    Returns:
        tuple: A tuple containing the relative path of the file and the line number of the
               matched frame. If no frame is matched, it returns a default key and -1.

    """
    regex = (
        r"(?<!_check_dataframe_schema_file)"
        r"(?<!_check_dataframe_schema)"
        r"(validate_dataframe_checkpoint|check_dataframe_schema)"
    )

    first_frames = stack[:7]
    first_frames.reverse()

    for i, frame in enumerate(first_frames):
        if frame.function == "wrapper" and i - 1 >= 0:
            next_frame = first_frames[i - 1]
            return _get_relative_path(next_frame.filename), next_frame.lineno

        if len(frame.code_context) >= 0 and re.search(regex, frame.code_context[0]):
            return _get_relative_path(frame.filename), frame.lineno
    return DEFAULT_KEY, -1


def _get_relative_path(file_path: str) -> str:
    """Get the relative path of a file.

    Args:
        file_path (str): The path to the file.

    Returns:
        str: The relative path of the file.

    """
    current_directory = os.getcwd()
    return os.path.relpath(file_path, current_directory)


def _update_validation_result(
    checkpoint_name: str, validation_status: str, output_path: Optional[str] = None
) -> None:
    """Update the validation result file with the status of a given checkpoint.

    Args:
        checkpoint_name (str): The name of the checkpoint to update.
        validation_status (str): The validation status to record for the checkpoint.
        output_path (str): The path to the output directory.

    Returns:
        None

    """
    _file = get_checkpoint_file(checkpoint_name)

    stack = inspect.stack()

    _file_from_stack, _line_of_code = _find_frame_in(stack)

    pipeline_result_metadata = ValidationResultsMetadata(output_path)

    pipeline_result_metadata.clean()

    pipeline_result_metadata.add_validation_result(
        ValidationResult(
            timestamp=datetime.now().isoformat(),
            file=_file if _file else _file_from_stack,
            line_of_code=_line_of_code,
            checkpoint_name=checkpoint_name,
            result=validation_status,
        )
    )

    pipeline_result_metadata.save()



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from functools import wraps
from typing import Callable, Optional, TypeVar

from typing_extensions import ParamSpec


P = ParamSpec("P")
R = TypeVar("R")


def log(
    _func: Optional[Callable[P, R]] = None,
    *,
    logger: Optional[logging.Logger] = None,
    log_args: bool = True,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Log the function call and any exceptions that occur.

    Args:
        _func: The function to log.
        logger: The logger to use for logging. If not provided, a logger will be created using the
                function's module name.
        log_args: Whether to log the arguments passed to the function.

    Returns:
        A decorator that logs the function call and any exceptions that occur.

    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            _logger = logging.getLogger(func.__module__) if logger is None else logger
            if log_args:
                args_repr = [repr(a) for a in args]
                kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
                formatted_args = ", ".join([*args_repr, *kwargs_repr])
                _logger.debug("%s called with args %s", func.__name__, formatted_args)
            try:
                return func(*args, **kwargs)
            except Exception:
                _logger.exception("An error occurred in %s", func.__name__)
                raise

        return wrapper

    # Handle the case where the decorator is used without parentheses
    if _func is None:
        return decorator
    return decorator(_func)



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import json
import logging
import os
import shutil

from typing import Optional

import pandas
import pandera as pa

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import DoubleType as SparkDoubleType
from pyspark.sql.types import StringType as SparkStringType
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    CHECKPOINT_JSON_OUTPUT_FILE_NAME_FORMAT,
    COLUMNS_KEY,
    DATAFRAME_CUSTOM_DATA_KEY,
    DATAFRAME_PANDERA_SCHEMA_KEY,
    DECIMAL_COLUMN_TYPE,
    DOT_PARQUET_EXTENSION,
    INTEGER_TYPE_COLLECTION,
    NULL_COLUMN_TYPE,
    PANDAS_LONG_TYPE,
    PANDAS_OBJECT_TYPE_COLLECTION,
    CheckpointMode,
)
from snowflake.snowpark_checkpoints_collector.collection_result.model import (
    CollectionPointResult,
    CollectionPointResultManager,
    CollectionResult,
)
from snowflake.snowpark_checkpoints_collector.column_collection import (
    ColumnCollectorManager,
)
from snowflake.snowpark_checkpoints_collector.column_pandera_checks import (
    PanderaColumnChecksManager,
)
from snowflake.snowpark_checkpoints_collector.snow_connection_model import (
    SnowConnection,
)
from snowflake.snowpark_checkpoints_collector.utils import (
    checkpoint_name_utils,
    file_utils,
)
from snowflake.snowpark_checkpoints_collector.utils.extra_config import (
    get_checkpoint_mode,
    get_checkpoint_sample,
    is_checkpoint_enabled,
)
from snowflake.snowpark_checkpoints_collector.utils.logging_utils import log
from snowflake.snowpark_checkpoints_collector.utils.telemetry import report_telemetry


LOGGER = logging.getLogger(__name__)


@log
def collect_dataframe_checkpoint(
    df: SparkDataFrame,
    checkpoint_name: str,
    sample: Optional[float] = None,
    mode: Optional[CheckpointMode] = None,
    output_path: Optional[str] = None,
) -> None:
    """Collect a DataFrame checkpoint.

    Args:
        df (SparkDataFrame): The input Spark DataFrame to collect.
        checkpoint_name (str): The name of the checkpoint.
        sample (float, optional): Fraction of DataFrame to sample for schema inference.
            Defaults to 1.0.
        mode (CheckpointMode): The mode to execution the collection.
            Defaults to CheckpointMode.Schema
        output_path (str, optional): The output path to save the checkpoint.
            Defaults to Current working Directory.

    Raises:
        Exception: Invalid mode value.
        Exception: Invalid checkpoint name. Checkpoint names must only contain alphanumeric characters
                   , underscores and dollar signs.

    """
    normalized_checkpoint_name = checkpoint_name_utils.normalize_checkpoint_name(
        checkpoint_name
    )
    if normalized_checkpoint_name != checkpoint_name:
        LOGGER.info(
            "Checkpoint name '%s' was normalized to '%s'",
            checkpoint_name,
            normalized_checkpoint_name,
        )
    is_valid_checkpoint_name = checkpoint_name_utils.is_valid_checkpoint_name(
        normalized_checkpoint_name
    )
    if not is_valid_checkpoint_name:
        raise Exception(
            f"Invalid checkpoint name: {normalized_checkpoint_name}. "
            "Checkpoint names must only contain alphanumeric characters, underscores and dollar signs."
        )
    if not is_checkpoint_enabled(normalized_checkpoint_name):
        LOGGER.info(
            "Checkpoint '%s' is disabled. Skipping collection.",
            normalized_checkpoint_name,
        )
        return

    LOGGER.info("Starting to collect checkpoint '%s'", normalized_checkpoint_name)
    LOGGER.debug("DataFrame size: %s rows", df.count())
    LOGGER.debug("DataFrame schema: %s", df.schema)

    collection_point_file_path = file_utils.get_collection_point_source_file_path()
    collection_point_line_of_code = file_utils.get_collection_point_line_of_code()
    collection_point_result = CollectionPointResult(
        collection_point_file_path,
        collection_point_line_of_code,
        normalized_checkpoint_name,
    )

    try:
        if _is_empty_dataframe_without_schema(df):
            raise Exception(
                "It is not possible to collect an empty DataFrame without schema"
            )

        _mode = get_checkpoint_mode(normalized_checkpoint_name, mode)

        if _mode == CheckpointMode.SCHEMA:
            column_type_dict = _get_spark_column_types(df)
            _sample = get_checkpoint_sample(normalized_checkpoint_name, sample)
            LOGGER.info(
                "Collecting checkpoint in %s mode using sample value %s",
                CheckpointMode.SCHEMA.name,
                _sample,
            )
            _collect_dataframe_checkpoint_mode_schema(
                normalized_checkpoint_name,
                df,
                _sample,
                column_type_dict,
                output_path,
            )
        elif _mode == CheckpointMode.DATAFRAME:
            LOGGER.info(
                "Collecting checkpoint in %s mode", CheckpointMode.DATAFRAME.name
            )
            snow_connection = SnowConnection()
            _collect_dataframe_checkpoint_mode_dataframe(
                normalized_checkpoint_name, df, snow_connection, output_path
            )
        else:
            raise Exception(f"Invalid mode value: {_mode}")

        collection_point_result.result = CollectionResult.PASS
        LOGGER.info(
            "Checkpoint '%s' collected successfully", normalized_checkpoint_name
        )

    except Exception as err:
        collection_point_result.result = CollectionResult.FAIL
        error_message = str(err)
        raise Exception(error_message) from err

    finally:
        collection_point_result_manager = CollectionPointResultManager(output_path)
        collection_point_result_manager.add_result(collection_point_result)


@report_telemetry(params_list=["column_type_dict"])
def _collect_dataframe_checkpoint_mode_schema(
    checkpoint_name: str,
    df: SparkDataFrame,
    sample: float,
    column_type_dict: dict[str, any],
    output_path: Optional[str] = None,
) -> None:
    sampled_df = df.sample(sample)
    if sampled_df.isEmpty():
        LOGGER.warning("Sampled DataFrame is empty. Collecting full DataFrame.")
        sampled_df = df

    pandas_df = _to_pandas(sampled_df)
    is_empty_df_with_object_column = _is_empty_dataframe_with_object_column(df)
    if is_empty_df_with_object_column:
        LOGGER.debug(
            "DataFrame is empty with object column. Skipping Pandera schema inference."
        )
        pandera_infer_schema = {}
    else:
        LOGGER.debug("Inferring Pandera schema from DataFrame")
        pandera_infer_schema = pa.infer_schema(pandas_df)

    column_name_collection = df.schema.names
    columns_to_remove_from_pandera_schema_collection = []
    column_custom_data_collection = []
    column_collector_manager = ColumnCollectorManager()
    column_pandera_checks_manager = PanderaColumnChecksManager()

    for column_name in column_name_collection:
        struct_field_column = column_type_dict[column_name]
        column_type = struct_field_column.dataType.typeName()
        LOGGER.info("Collecting column '%s' of type '%s'", column_name, column_type)
        pyspark_column = df.select(col(column_name))

        is_column_to_remove_from_pandera_schema = (
            _is_column_to_remove_from_pandera_schema(column_type)
        )
        if is_column_to_remove_from_pandera_schema:
            columns_to_remove_from_pandera_schema_collection.append(column_name)

        is_empty_column = (
            pyspark_column.dropna().isEmpty() and column_type is not NULL_COLUMN_TYPE
        )
        if is_empty_column:
            LOGGER.debug("Column '%s' is empty.", column_name)
            custom_data = column_collector_manager.collect_empty_custom_data(
                column_name, struct_field_column, pyspark_column
            )
            column_custom_data_collection.append(custom_data)
            continue

        pandera_column = pandera_infer_schema.columns[column_name]
        pandera_column.checks = []
        column_pandera_checks_manager.add_checks_column(
            column_name, column_type, df, pandera_column
        )

        custom_data = column_collector_manager.collect_column(
            column_name, struct_field_column, pyspark_column
        )
        column_custom_data_collection.append(custom_data)

    pandera_infer_schema_dict = _get_pandera_infer_schema_as_dict(
        pandera_infer_schema,
        is_empty_df_with_object_column,
        columns_to_remove_from_pandera_schema_collection,
    )

    dataframe_custom_column_data = {COLUMNS_KEY: column_custom_data_collection}
    dataframe_schema_contract = {
        DATAFRAME_PANDERA_SCHEMA_KEY: pandera_infer_schema_dict,
        DATAFRAME_CUSTOM_DATA_KEY: dataframe_custom_column_data,
    }

    dataframe_schema_contract_json = json.dumps(dataframe_schema_contract)
    _generate_json_checkpoint_file(
        checkpoint_name, dataframe_schema_contract_json, output_path
    )


def _get_spark_column_types(df: SparkDataFrame) -> dict[str, StructField]:
    schema = df.schema
    column_type_collection = {}
    for field in schema.fields:
        column_name = field.name
        column_type_collection[column_name] = field
    return column_type_collection


def _is_empty_dataframe_without_schema(df: SparkDataFrame) -> bool:
    is_empty = df.isEmpty()
    has_schema = len(df.schema.fields) > 0
    return is_empty and not has_schema


def _is_empty_dataframe_with_object_column(df: SparkDataFrame):
    is_empty = df.isEmpty()
    if not is_empty:
        return False

    for field in df.schema.fields:
        if field.dataType.typeName() in PANDAS_OBJECT_TYPE_COLLECTION:
            return True

    return False


def _is_column_to_remove_from_pandera_schema(column_type) -> bool:
    is_decimal_type = column_type == DECIMAL_COLUMN_TYPE
    return is_decimal_type


def _get_pandera_infer_schema_as_dict(
    pandera_infer_schema, is_empty_df_with_string_column, columns_to_remove_collection
) -> dict[str, any]:
    if is_empty_df_with_string_column:
        return {}

    pandera_infer_schema_dict = json.loads(pandera_infer_schema.to_json())
    for column in columns_to_remove_collection:
        LOGGER.debug("Removing column '%s' from Pandera schema", column)
        del pandera_infer_schema_dict[COLUMNS_KEY][column]

    return pandera_infer_schema_dict


def _generate_json_checkpoint_file(
    checkpoint_name, dataframe_schema_contract, output_path: Optional[str] = None
) -> None:
    checkpoint_file_name = CHECKPOINT_JSON_OUTPUT_FILE_NAME_FORMAT.format(
        checkpoint_name
    )
    output_directory_path = file_utils.get_output_directory_path(output_path)
    checkpoint_file_path = os.path.join(output_directory_path, checkpoint_file_name)
    LOGGER.info("Writing DataFrame JSON schema file to '%s'", checkpoint_file_path)
    with open(checkpoint_file_path, "w") as f:
        f.write(dataframe_schema_contract)


@report_telemetry(params_list=["df"])
def _collect_dataframe_checkpoint_mode_dataframe(
    checkpoint_name: str,
    df: SparkDataFrame,
    snow_connection: SnowConnection,
    output_path: Optional[str] = None,
) -> None:
    output_path = file_utils.get_output_directory_path(output_path)
    parquet_directory = os.path.join(output_path, checkpoint_name)
    generate_parquet_for_spark_df(df, parquet_directory)
    _create_snowflake_table_from_parquet(
        checkpoint_name, parquet_directory, snow_connection
    )


def generate_parquet_for_spark_df(spark_df: SparkDataFrame, output_path: str) -> None:
    """Generate a parquet file from a Spark DataFrame.

    This function will  convert Float to Double to avoid precision problems.
    Spark parquet use IEEE 32-bit floating point values,
    while Snowflake uses IEEE 64-bit floating point values.

    Args:
        spark_df: dataframe to be saved as parquet
        output_path: path to save the parquet files.
    returns: None

    Raises:
        Exception: No parquet files were generated.

    """
    new_cols = [
        (
            col(c).cast(SparkStringType()).cast(SparkDoubleType()).alias(c)
            if t == "float"
            else col(c)
        )
        for (c, t) in spark_df.dtypes
    ]
    converted_df = spark_df.select(new_cols)

    if os.path.exists(output_path):
        LOGGER.warning(
            "Output directory '%s' already exists. Deleting it...", output_path
        )
        shutil.rmtree(output_path)

    LOGGER.info("Writing DataFrame to parquet files at '%s'", output_path)
    converted_df.write.parquet(output_path, mode="overwrite")

    target_dir = os.path.join(output_path, "**", f"*{DOT_PARQUET_EXTENSION}")
    parquet_files = glob.glob(target_dir, recursive=True)
    parquet_files_count = len(parquet_files)
    if parquet_files_count == 0:
        raise Exception("No parquet files were generated.")
    LOGGER.info(
        "%s parquet files were written in '%s'",
        parquet_files_count,
        output_path,
    )


def _create_snowflake_table_from_parquet(
    table_name: str, input_path: str, snow_connection: SnowConnection
) -> None:
    snow_connection.create_snowflake_table_from_local_parquet(table_name, input_path)


def _to_pandas(sampled_df: SparkDataFrame) -> pandas.DataFrame:
    LOGGER.debug("Converting Spark DataFrame to Pandas DataFrame")
    pandas_df = sampled_df.toPandas()
    for field in sampled_df.schema.fields:
        has_nan = pandas_df[field.name].isna().any()
        is_integer = field.dataType.typeName() in INTEGER_TYPE_COLLECTION
        if has_nan and is_integer:
            LOGGER.debug(
                "Converting column '%s' to '%s' type",
                field.name,
                PANDAS_LONG_TYPE,
            )
            pandas_df[field.name] = pandas_df[field.name].astype(PANDAS_LONG_TYPE)

    return pandas_df



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


# Add a NullHandler to prevent logging messages from being output to
# sys.stderr if no logging configuration is provided.
logging.getLogger(__name__).addHandler(logging.NullHandler())

# ruff: noqa: E402

from snowflake.hypothesis_snowpark.strategies import dataframe_strategy


__all__ = [
    "dataframe_strategy",
]



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import random
import re

from contextlib import contextmanager
from typing import Optional, Union

import numpy as np
import pandas as pd
import pandera as pa

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_COLUMNS_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_ROWS_COUNT,
    CUSTOM_DATA_ROWS_NULL_COUNT_KEY,
    CUSTOM_DATA_TYPE_KEY,
)
from snowflake.hypothesis_snowpark.datatype_mapper import (
    pandera_dtype_to_snowpark_dtype,
    pyspark_dtype_to_snowpark_dtype,
)
from snowflake.snowpark import DataFrame, Session
from snowflake.snowpark.types import StructField, StructType


LOGGER = logging.getLogger(__name__)


def load_json_schema(json_schema: str) -> dict:
    """Load the JSON schema from a file.

    Args:
        json_schema: The path to the JSON file.

    Returns:
        The JSON schema as a dictionary.

    """
    if not os.path.isfile(json_schema):
        raise ValueError(f"Invalid JSON schema path: {json_schema}")

    try:
        with open(json_schema, encoding="utf-8") as file:
            return json.load(file)
    except (OSError, json.JSONDecodeError) as e:
        raise ValueError(f"Error reading JSON schema file: {e}") from None


def generate_snowpark_dataframe(
    pandas_df: pd.DataFrame,
    snowpark_schema: StructType,
    session: Session,
) -> DataFrame:
    """Generate a Snowpark DataFrame from a Pandas DataFrame.

    Args:
        pandas_df: The Pandas DataFrame to convert.
        snowpark_schema: The Snowpark schema to use for creating the DataFrame.
        session: The Snowpark session to use for creating the DataFrame.

    Returns:
        A Snowpark DataFrame.

    """
    pandas_df = pandas_df.replace({np.nan: None})

    # Snowpark ignores the schema argument if the data argument is a pandas dataframe. We need to convert the pandas
    # dataframe into a list of tuples to be able to specify a schema.
    data = list(pandas_df.itertuples(index=False, name=None))

    # In some scenarios, the create_dataframe function performs some random operations (e.g., generating temporary table
    # names or temporary stage names). Due to Hypothesis controls randomness during its tests, we need to temporarily
    # set a random seed to ensure that the names generated by Snowpark are unique to avoid conflicts.
    with temporary_random_seed():
        LOGGER.info(
            "Creating Snowpark DataFrame with schema=%s and size=%s",
            snowpark_schema,
            len(data),
        )
        snowpark_df = session.create_dataframe(data=data, schema=snowpark_schema)
    return snowpark_df


def apply_custom_null_values(
    pandas_df: pd.DataFrame, custom_data: dict
) -> pd.DataFrame:
    """Apply null values to a Pandas DataFrame based on the custom data.

    This function modifies the DataFrame in place and returns a new DataFrame with null values applied.

    Args:
        pandas_df: The Pandas DataFrame to apply null values to.
        custom_data: The custom data associated with the Pandera schema.

    Returns:
        A Pandas DataFrame with null values applied.

    """
    null_proportions = {}
    for col in custom_data.get(CUSTOM_DATA_COLUMNS_KEY, []):
        rows_count = col.get(CUSTOM_DATA_ROWS_COUNT, 0)
        if rows_count > 0:
            null_proportions[col[CUSTOM_DATA_NAME_KEY]] = (
                col[CUSTOM_DATA_ROWS_NULL_COUNT_KEY] / rows_count
            )

    df_copy = pandas_df.copy()
    total_rows = len(df_copy)

    for column, target_null_proportion in null_proportions.items():
        if column not in df_copy.columns:
            LOGGER.warning(
                "Column '%s' not found in DataFrame during null value correction. Skipping it.",
                column,
            )
            continue

        required_nulls = round(total_rows * target_null_proportion)
        current_nulls = df_copy[column].isnull().sum()

        if current_nulls < required_nulls:
            LOGGER.debug(
                "Column '%s' has %s null values but %s are required.",
                column,
                current_nulls,
                required_nulls,
            )
            additional_nulls = required_nulls - current_nulls
            non_null_positions = np.where(df_copy[column].notnull())[0]
            selected_positions = np.random.choice(
                non_null_positions, size=additional_nulls, replace=False
            )
            df_copy[column] = df_copy[column].astype(object)
            df_copy.iloc[selected_positions, df_copy.columns.get_loc(column)] = None
            LOGGER.info(
                "Added %s null values to column '%s'",
                additional_nulls,
                column,
            )

    return df_copy.replace({np.nan: None})


@contextmanager
def temporary_random_seed(seed: Union[int, float, str, bytes, bytearray, None] = None):
    """Set a temporary random seed within a context.

    This context manager saves the current state of the random number generator,
    sets a new seed, and then restores the original state when exiting the context.

    Args:
        seed: The seed to set for the random number generator. If not provided, seeds from current time or from an
              operating system specific randomness source if available.

    Yields:
        None: The context manager does not return anything.

    """
    current_state = random.getstate()
    random.seed(seed)
    try:
        yield
    finally:
        random.setstate(current_state)


def generate_snowpark_schema(
    pandera_schema: pa.DataFrameSchema, custom_data: Optional[dict] = None
) -> StructType:
    """Generate a Snowpark schema from a Pandera schema.

    Args:
        pandera_schema: The Pandera schema to convert.
        custom_data: The custom data associated with the Pandera schema.

    Returns:
        A Snowpark struct type.

    """
    struct_fields = []

    if custom_data is None:
        for column_name, column_data in pandera_schema.columns.items():
            snowpark_dtype = pandera_dtype_to_snowpark_dtype(column_data.dtype)
            struct_field = StructField(
                column_identifier=column_name,
                datatype=snowpark_dtype,
                nullable=column_data.nullable,
            )
            struct_fields.append(struct_field)
    else:
        for column_name, column_data in pandera_schema.columns.items():
            custom_data_column = next(
                (
                    column
                    for column in custom_data.get(CUSTOM_DATA_COLUMNS_KEY, [])
                    if column.get(CUSTOM_DATA_NAME_KEY) == column_name
                ),
                None,
            )

            if custom_data_column is None:
                raise ValueError(f"Column '{column_name}' is missing from custom_data")

            pyspark_type = custom_data_column.get(CUSTOM_DATA_TYPE_KEY)

            if pyspark_type is None:
                raise ValueError(
                    f"Type for column '{column_name}' is missing from custom_data"
                )

            snowpark_type = pyspark_dtype_to_snowpark_dtype(pyspark_type)

            struct_field = StructField(
                column_identifier=column_name,
                datatype=snowpark_type,
                nullable=column_data.nullable,
            )
            struct_fields.append(struct_field)

    return StructType(struct_fields)


def replace_surrogate_chars(df: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    """Replace surrogate characters in the specified columns of a Pandas DataFrame with a space.

    Args:
        df: The Pandas DataFrame to process.
        columns: The list of columns to process.

    Returns:
        A new DataFrame with surrogate characters replaced.

    """
    surrogate_pattern = re.compile(r"[\uD800-\uDFFF]")
    replacement = " "
    df_copy = df.copy()

    for column in columns:
        if column not in df_copy.columns:
            LOGGER.warning(
                "Column '%s' not found in DataFrame during surrogate replacement. Skipping it.",
                column,
            )
            continue

        LOGGER.info("Replacing surrogate characters in column '%s'", column)
        df_copy[column] = df_copy[column].apply(
            lambda value: re.sub(surrogate_pattern, replacement, value)
            if isinstance(value, str)
            else value
        )

    return df_copy



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from typing import Optional

from snowflake.snowpark_checkpoints_configuration.model.checkpoints import (
    Checkpoint,
    Checkpoints,
)
from snowflake.snowpark_checkpoints_configuration.singleton import Singleton


LOGGER = logging.getLogger(__name__)


class CheckpointMetadata(metaclass=Singleton):

    """CheckpointMetadata class.

    This is a singleton class that reads the checkpoints.json file
    and provides an interface to get the checkpoint configuration.

    Args:
        metaclass (Singleton, optional): Defaults to Singleton.

    """

    def __init__(self, path: Optional[str] = None):
        self.checkpoint_model: Checkpoints = Checkpoints(type="", pipelines=[])
        directory = path if path is not None else os.getcwd()
        checkpoints_file = os.path.join(directory, "checkpoints.json")
        if os.path.exists(checkpoints_file):
            LOGGER.info("Reading checkpoints file: '%s'", checkpoints_file)
            try:
                with open(checkpoints_file) as f:
                    checkpoint_json = f.read()
                    self.checkpoint_model = Checkpoints.model_validate_json(
                        checkpoint_json
                    )
                LOGGER.info(
                    "Successfully read and validated checkpoints file: '%s'",
                    checkpoints_file,
                )
            except Exception as e:
                error_msg = f"An error occurred while reading the checkpoints file: '{checkpoints_file}'"
                LOGGER.exception(error_msg)
                raise Exception(f"{error_msg} \n {e}") from None
        else:
            LOGGER.warning("Checkpoints file not found: '%s'", checkpoints_file)

    def get_checkpoint(self, checkpoint_name: str) -> Checkpoint:
        """Get a checkpoint by its name.

        Args:
            checkpoint_name (str): checkpoint name

        Returns:
            Checkpoint: Checkpoint configuration instance

        """
        return self.checkpoint_model.get_check_point(checkpoint_name=checkpoint_name)



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pandas as pd

from pandera import Check, Column
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import col as spark_col
from pyspark.sql.functions import length as spark_length
from pyspark.sql.functions import max as spark_max
from pyspark.sql.functions import min as spark_min

from snowflake.snowpark_checkpoints_collector.collection_common import (
    BETWEEN_CHECK_ERROR_MESSAGE_FORMAT,
    BOOLEAN_COLUMN_TYPE,
    BYTE_COLUMN_TYPE,
    COLUMN_MAX_KEY,
    COLUMN_MIN_KEY,
    DAYTIMEINTERVAL_COLUMN_TYPE,
    DOUBLE_COLUMN_TYPE,
    FLOAT_COLUMN_TYPE,
    INTEGER_COLUMN_TYPE,
    LONG_COLUMN_TYPE,
    SHORT_COLUMN_TYPE,
    STRING_COLUMN_TYPE,
    TIMESTAMP_COLUMN_TYPE,
    TIMESTAMP_NTZ_COLUMN_TYPE,
)


LOGGER = logging.getLogger(__name__)


def collector_register(cls):
    """Decorate a class with the checks mechanism.

    Args:
        cls: The class to decorate.

    Returns:
        The class to decorate.

    """
    LOGGER.debug("Starting to register checks from class %s", cls.__name__)
    cls._collectors = {}
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        if hasattr(method, "_column_type"):
            col_type_collection = method._column_type
            for col_type in col_type_collection:
                cls._collectors[col_type] = method_name
                LOGGER.debug(
                    "Registered check '%s' for column type '%s'", method_name, col_type
                )
    return cls


def column_register(*args):
    """Decorate a method to register it in the checks mechanism based on column type.

    Args:
        args: the column type to register.

    Returns:
        The wrapper.

    """

    def wrapper(func):
        has_arguments = len(args) > 0
        if has_arguments:
            func._column_type = args
        return func

    return wrapper


@collector_register
class PanderaColumnChecksManager:

    """Manage class for Pandera column checks based on type."""

    def add_checks_column(
        self,
        clm_name: str,
        clm_type: str,
        pyspark_df: SparkDataFrame,
        pandera_column: Column,
    ) -> None:
        """Add checks to Pandera column based on the column type.

        Args:
            clm_name (str): the name of the column.
            clm_type (str): the type of the column.
            pyspark_df (pyspark.sql.DataFrame): the DataFrame.
            pandera_column (pandera.Column): the Pandera column.

        """
        if clm_type not in self._collectors:
            LOGGER.debug(
                "No Pandera checks found for column '%s' of type '%s'. Skipping checks for this column.",
                clm_name,
                clm_type,
            )
            return

        func_name = self._collectors[clm_type]
        func = getattr(self, func_name)
        LOGGER.debug(
            "Adding Pandera checks to column '%s' of type '%s'", clm_name, clm_type
        )
        func(clm_name, pyspark_df, pandera_column)

    @column_register(BOOLEAN_COLUMN_TYPE)
    def _add_boolean_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        pandera_column.checks.extend([Check.isin([True, False])])

    @column_register(DAYTIMEINTERVAL_COLUMN_TYPE)
    def _add_daytimeinterval_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        select_result = pyspark_df.select(
            spark_min(spark_col(clm_name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(clm_name)).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_value = pd.to_timedelta(select_result[COLUMN_MIN_KEY])
        max_value = pd.to_timedelta(select_result[COLUMN_MAX_KEY])

        pandera_column.checks.append(
            Check.between(
                min_value=min_value,
                max_value=max_value,
                include_max=True,
                include_min=True,
                title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
            )
        )

    @column_register(
        BYTE_COLUMN_TYPE,
        SHORT_COLUMN_TYPE,
        INTEGER_COLUMN_TYPE,
        LONG_COLUMN_TYPE,
        FLOAT_COLUMN_TYPE,
        DOUBLE_COLUMN_TYPE,
    )
    def _add_numeric_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        select_result = pyspark_df.select(
            spark_min(spark_col(clm_name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(clm_name)).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_value = select_result[COLUMN_MIN_KEY]
        max_value = select_result[COLUMN_MAX_KEY]

        pandera_column.checks.append(
            Check.between(
                min_value=min_value,
                max_value=max_value,
                include_max=True,
                include_min=True,
                title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
            )
        )

    @column_register(STRING_COLUMN_TYPE)
    def _add_string_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        select_result = pyspark_df.select(
            spark_min(spark_length(spark_col(clm_name))).alias(COLUMN_MIN_KEY),
            spark_max(spark_length(spark_col(clm_name))).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_length = select_result[COLUMN_MIN_KEY]
        max_length = select_result[COLUMN_MAX_KEY]

        pandera_column.checks.append(Check.str_length(min_length, max_length))

    @column_register(TIMESTAMP_COLUMN_TYPE)
    def _add_timestamp_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        select_result = pyspark_df.select(
            spark_min(spark_col(clm_name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(clm_name)).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_value = pd.Timestamp(select_result[COLUMN_MIN_KEY])
        max_value = pd.Timestamp(select_result[COLUMN_MAX_KEY])

        pandera_column.checks.append(
            Check.between(
                min_value=min_value,
                max_value=max_value,
                include_max=True,
                include_min=True,
                title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
            )
        )

    @column_register(TIMESTAMP_NTZ_COLUMN_TYPE)
    def _add_timestamp_ntz_type_checks(
        self, clm_name: str, pyspark_df: SparkDataFrame, pandera_column: Column
    ) -> None:
        select_result = pyspark_df.select(
            spark_min(spark_col(clm_name)).alias(COLUMN_MIN_KEY),
            spark_max(spark_col(clm_name)).alias(COLUMN_MAX_KEY),
        ).collect()[0]

        min_value = pd.Timestamp(select_result[COLUMN_MIN_KEY])
        max_value = pd.Timestamp(select_result[COLUMN_MAX_KEY])

        pandera_column.checks.append(
            Check.between(
                min_value=min_value,
                max_value=max_value,
                include_max=True,
                include_min=True,
                title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
            )
        )



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from datetime import datetime
from typing import Optional

import pandas as pd

from pyspark.sql import SparkSession

from snowflake.snowpark import Session
from snowflake.snowpark_checkpoints.utils.constants import SCHEMA_EXECUTION_MODE


LOGGER = logging.getLogger(__name__)
RESULTS_TABLE = "SNOWPARK_CHECKPOINTS_REPORT"


class SnowparkJobContext:

    """Class used to record migration results in Snowflake.

    Args:
        snowpark_session: A Snowpark session instance.
        spark_session: A Spark session instance.
        job_name: The name of the job.
        log_results: Whether to log the migration results in Snowflake.

    """

    def __init__(
        self,
        snowpark_session: Session,
        spark_session: SparkSession = None,
        job_name: Optional[str] = None,
        log_results: Optional[bool] = True,
    ):
        self.log_results = log_results
        self.job_name = job_name
        self.spark_session = spark_session or self._create_pyspark_session()
        self.snowpark_session = snowpark_session

    def _mark_fail(
        self, message, checkpoint_name, data, execution_mode=SCHEMA_EXECUTION_MODE
    ):
        if not self.log_results:
            LOGGER.warning(
                (
                    "Recording of migration results into Snowflake is disabled. "
                    "Failure result for checkpoint '%s' will not be recorded."
                ),
                checkpoint_name,
            )
            return

        LOGGER.debug(
            "Marking failure for checkpoint '%s' in '%s' mode with message '%s'",
            checkpoint_name,
            execution_mode,
            message,
        )

        session = self.snowpark_session
        df = pd.DataFrame(
            {
                "DATE": [datetime.now()],
                "JOB": [self.job_name],
                "STATUS": ["fail"],
                "CHECKPOINT": [checkpoint_name],
                "MESSAGE": [message],
                "DATA": [f"{data}"],
                "EXECUTION_MODE": [execution_mode],
            }
        )
        report_df = session.createDataFrame(df)
        LOGGER.info("Writing failure result to table: '%s'", RESULTS_TABLE)
        report_df.write.mode("append").save_as_table(RESULTS_TABLE)

    def _mark_pass(self, checkpoint_name, execution_mode=SCHEMA_EXECUTION_MODE):
        if not self.log_results:
            LOGGER.warning(
                (
                    "Recording of migration results into Snowflake is disabled. "
                    "Pass result for checkpoint '%s' will not be recorded."
                ),
                checkpoint_name,
            )
            return

        LOGGER.debug(
            "Marking pass for checkpoint '%s' in '%s' mode",
            checkpoint_name,
            execution_mode,
        )

        session = self.snowpark_session
        df = pd.DataFrame(
            {
                "DATE": [datetime.now()],
                "JOB": [self.job_name],
                "STATUS": ["pass"],
                "CHECKPOINT": [checkpoint_name],
                "MESSAGE": [""],
                "DATA": [""],
                "EXECUTION_MODE": [execution_mode],
            }
        )
        report_df = session.createDataFrame(df)
        LOGGER.info("Writing pass result to table: '%s'", RESULTS_TABLE)
        report_df.write.mode("append").save_as_table(RESULTS_TABLE)

    def _create_pyspark_session(self) -> SparkSession:
        LOGGER.info("Creating a PySpark session")
        return SparkSession.builder.getOrCreate()



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext


class SnowparkCheckpointError(Exception):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name: str,
        data=None,
    ):
        job_name = job_context.job_name if job_context else "Unknown Job"
        super().__init__(
            f"Job: {job_name} Checkpoint: {checkpoint_name}\n{message} \n {data}"
        )
        if job_context:
            job_context._mark_fail(
                message,
                checkpoint_name,
                data,
            )


class SparkMigrationError(SnowparkCheckpointError):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name=None,
        data=None,
    ):
        super().__init__(message, job_context, checkpoint_name, data)


class SchemaValidationError(SnowparkCheckpointError):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name=None,
        data=None,
    ):
        super().__init__(message, job_context, checkpoint_name, data)
